from argparse import Namespace
from pathlib import Path
from typing import Iterable


def load_scope(args: Namespace, config: dict) -> Iterable[Path]:
    """Load a list of files in scope based on args and config"""
    include_files = args.files or config.include.files
    exclude_files = args.xfiles or config.exclude.files
    src = Path(args.src)

    if src.is_file():
        yield src

    else:
        for include in include_files:
            for filepath in src.rglob(include):
                if exclude_files and exclude_files.match(filepath.as_posix()):
                    continue

                yield filepath



import json
import re
from pathlib import Path
from typing import Iterator

from whispers.core.utils import global_exception_handler
from whispers.models.pair import KeyValuePair
from whispers.plugins.traverse import StructuredDocument


class Json(StructuredDocument):
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        """
        Convert custom JSON to parsable JSON
        - Remove lines that start with // comments
        - Strip // comments from the end the line
        """
        document = ""
        for line in filepath.open("r").readlines():
            if line.startswith("//"):
                continue
            line = re.sub(r" // ?.*$", "", line)
            document += line

        # Load converted JSON
        try:
            document = json.loads(document)
            yield from self.traverse(document)
        except Exception:
            global_exception_handler(filepath.as_posix(), document)



from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class KeyValuePair:
    """Main key-value pair class"""

    key: str
    value: str
    keypath: List = field(default_factory=list)
    file: str = ""
    line: int = 0
    rule: object = None

    def __post_init__(self) -> None:
        if self.keypath == []:
            self.keypath = [self.key]

    def __repr__(self) -> str:
        return f"[{self.file}:{self.line}] {self.key} = {self.value}"

    def to_json(self) -> Dict:
        return {
            "key": self.key,
            "value": self.value,
            "file": self.file,
            "line": self.line,
            "rule_id": self.rule.id,
            "message": self.rule.message,
            "severity": self.rule.severity,
        }



import re
from pathlib import Path
from typing import Iterator

import yaml
from yaml.parser import ParserError
from yaml.resolver import Resolver

from whispers.core.utils import global_exception_handler
from whispers.models.pair import KeyValuePair
from whispers.plugins.traverse import StructuredDocument


class Yml(StructuredDocument):
    def __init__(self):
        super().__init__()

        # Remove resolvers for on/off/yes/no
        list(map(lambda idx: Resolver.yaml_implicit_resolvers.pop(idx, None), "OoYyNn"))

    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        def _constructor(loader, tag_suffix, node):
            """This is needed to parse IaC syntax"""
            ret = loader.construct_scalar(node)
            return f"{tag_suffix} {ret}"

        """
        Convert custom YAML to parsable YAML
        - Skip ---
        - Quote unquoted values such as {{ placeholder }}
        - Remove text between <% %> and {% %}
        - Remove comments that start with #
        """
        regex_unquoted = re.compile(r".+(\[)?\{\{.*\}\}(\])?")
        document = ""

        for line in filepath.open("r").readlines():
            if line.startswith("---"):
                continue

            if regex_unquoted.match(line):
                line = line.replace("{{", "'{{").replace("}}", "}}'")

            document += line

        document = re.sub(r"[<{]%.*?%[}>]", "", document, flags=re.MULTILINE | re.DOTALL)
        document = re.sub(r"^#.*$", "", document)

        # Load converted YAML
        try:
            yaml.add_multi_constructor("", _constructor, Loader=yaml.SafeLoader)
            code = yaml.safe_load(document)
            yield from self.traverse(code)

        except ParserError:
            global_exception_handler(filepath.as_posix(), document)



from configparser import ConfigParser, MissingSectionHeaderError
from itertools import chain
from pathlib import Path
from typing import Iterator, Optional

from crossplane import parse as nginx_parse

from whispers.core.utils import strip_string
from whispers.models.pair import KeyValuePair
from whispers.plugins.common import Common
from whispers.plugins.traverse import StructuredDocument
from whispers.plugins.xml import Xml


class Config:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        if "<?xml " in filepath.open("r").readline():
            yield from Xml().pairs(filepath)

        else:
            # Attempt to parse as Nginx config
            yield from self.parse_as_nginx(filepath)

            try:
                # Attempt to parse as Windows INI
                yield from self.parse_as_ini(filepath)

            except MissingSectionHeaderError:
                # Otherwise, parse as text line by line
                yield from self.parse_as_text(filepath)

    @staticmethod
    def parse_as_ini(filepath: Path) -> Optional[KeyValuePair]:
        """Parse file as Windows .ini"""
        parser = ConfigParser()
        parser.read(filepath.as_posix())
        sections = map(lambda section: section.items(), parser.values())
        items = chain.from_iterable(sections)

        for item in items:
            key, value = item
            yield KeyValuePair(key, value)

    @staticmethod
    def parse_as_text(filepath: Path) -> Optional[KeyValuePair]:
        """Parse as plain text"""
        for lineno, line in enumerate(filepath.open(), 1):
            line = line.strip()
            if "=" not in line:
                continue

            key, value = line.split("=", 1)
            key = strip_string(key)
            value = strip_string(value)

            if value:
                yield KeyValuePair(key, value, line=lineno)

    @staticmethod
    def parse_as_nginx(filepath: Path) -> Optional[Iterator[KeyValuePair]]:
        """Parse file as nginx.conf"""
        nginx_conf = nginx_parse(filepath, strict=False, single=True)
        if nginx_conf.get("status") != "ok":
            return  # skip failed (ie: not nginx.conf format)

        pairs = StructuredDocument().traverse(nginx_conf)
        args = filter(lambda pair: pair.key == "args", pairs)

        while True:
            try:
                key = next(args).value
                value = next(args).value
                keypath = [key, value]
                yield KeyValuePair(key, value, keypath, filepath.as_posix())
                yield from Common(keypath).parse_uri(value)

            except Exception:
                return



from pathlib import Path
from typing import Iterator

from whispers.core.utils import strip_string
from whispers.models.pair import KeyValuePair


class Htpasswd:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            if ":" not in line:
                continue

            creds = line.strip().split(":")
            value = strip_string(creds[1])
            if value:
                key = "htpasswd hash"
                yield KeyValuePair(key, value, line=lineno)



from pathlib import Path
from typing import Iterator
from urllib.parse import urlparse

from whispers.models.pair import KeyValuePair


class Pip:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            if "http" not in line:
                continue

            value = urlparse(line.split("=")[-1].strip()).password
            if value:
                key = "pip password"
                yield KeyValuePair(key, value, line=lineno)
